#![no_std]
#![no_main]

use cortex_m_rt::entry;

use rtt_target::{rprintln, rtt_init_print};
use stm32h7::stm32h753 as pac;

use panic_rtt_target as _;

// #[link_section = ".extflashdata"]
static TEST: [u8; 4] = [0xC0, 0xFE, 0xBA, 0xBE];

#[entry]
fn main() -> ! {
    rtt_init_print!();
    let mut dp = unsafe { pac::Peripherals::steal() };

    rprintln!("Configure ...");
    rcc(&mut dp);
    gpio_alt(&mut dp);
    fmc(&mut dp);

    let addr = nor::NOR_ADDR as *mut u16;

    // Get meta information from the flash to verify, that the communication with the flash works
    rprintln!("Command Mode: {:x}", nor::get_command_mode(addr));
    nor::get_status(addr).unwrap();
    rprintln!("{:x?}", nor::read_id(addr));

    let addr = nor::NOR_ADDR as *const u16;

    // Read some data from the memory mapped flash to verify, that data is in the flash
    unsafe {
        let v = core::ptr::read_volatile(addr);
        cortex_m::asm::dsb();
        rprintln!("Read from NORFLASH: {:x}", v);
    }

    loop {
        // Toggle LED to verify that the loop is running
        dp.GPIOF
            .odr
            .modify(|r, w| w.odr10().bit(r.odr10().is_low()));

        // Read test data out of external flash (if linker section is specified) and to force
        // JFlash to try flash to external flash.
        rprintln!("{:x?}", &TEST);
        cortex_m::asm::delay(1_100_000_000);
    }
}

#[inline(never)]
fn gpio_alt(dp: &mut pac::Peripherals) {
    unsafe {
        dp.GPIOD.moder.write(|w| w.bits(0xAAAA_AAFA));
        dp.GPIOD.ospeedr.write(|w| w.bits(0xFFFF_FF0F));
        dp.GPIOD.pupdr.write(|w| w.bits(0x5555_5505));
        dp.GPIOD.afrl.write(|w| w.bits(0xCCCC_00CC));
        dp.GPIOD.afrh.write(|w| w.bits(0xCCCC_CCCC));

        dp.GPIOE.moder.write(|w| w.bits(0xAAAA_AAAA));
        dp.GPIOE.ospeedr.write(|w| w.bits(0xFFFF_FFFF));
        dp.GPIOE.pupdr.write(|w| w.bits(0x5555_5555));
        dp.GPIOE.afrl.write(|w| w.bits(0xCCCC_CCCC));
        dp.GPIOE.afrh.write(|w| w.bits(0xCCCC_CCCC));

        dp.GPIOF.moder.write(|w| w.bits(0xAA9F_FAAA));
        dp.GPIOF.otyper.write(|w| w.bits(0x0000_0400));
        dp.GPIOF.ospeedr.write(|w| w.bits(0xFFF0_0FFF));
        dp.GPIOF.pupdr.write(|w| w.bits(0x5540_0555));
        dp.GPIOF.odr.write(|w| w.bits(0x0000_0400));
        dp.GPIOF.afrl.write(|w| w.bits(0x00CC_CCCC));
        dp.GPIOF.afrh.write(|w| w.bits(0xCCCC_C000));

        dp.GPIOG.moder.write(|w| w.bits(0xBFEE_FAAA));
        dp.GPIOG.ospeedr.write(|w| w.bits(0xC033_0FFF));
        dp.GPIOG.pupdr.write(|w| w.bits(0x4011_0555));
        dp.GPIOG.afrl.write(|w| w.bits(0x00CC_CCCC));
        dp.GPIOG.afrh.write(|w| w.bits(0xC000_0C0C));

        dp.GPIOH.moder.write(|w| w.bits(0xAAAA_ABFF));
        dp.GPIOH.ospeedr.write(|w| w.bits(0xFFFF_FC00));
        dp.GPIOH.pupdr.write(|w| w.bits(0x5555_5400));
        dp.GPIOH.afrl.write(|w| w.bits(0xCCC0_0000));
        dp.GPIOH.afrh.write(|w| w.bits(0xCCCC_CCCC));

        dp.GPIOI.moder.write(|w| w.bits(0xAAE8_AAAA));
        dp.GPIOI.ospeedr.write(|w| w.bits(0xAA3C_FFFF));
        dp.GPIOI.pupdr.write(|w| w.bits(0x0014_5555));
        dp.GPIOI.afrl.write(|w| w.bits(0xCCCC_CCCC));
        dp.GPIOI.afrh.write(|w| w.bits(0xEEEE_0CC0));
    }
}

#[inline(never)]
fn rcc(dp: &mut pac::Peripherals) {
    dp.RCC.ahb3enr.write(|w| w.fmcen().enabled());
    dp.RCC.c1_ahb3enr.write(|w| w.fmcen().enabled());
    dp.RCC.ahb4enr.write(|w| {
        w.gpioden()
            .enabled()
            .gpioden()
            .enabled()
            .gpioeen()
            .enabled()
            .gpiofen()
            .enabled()
            .gpiogen()
            .enabled()
            .gpiohen()
            .enabled()
            .gpioien()
            .enabled()
    });
}

#[inline(never)]
fn fmc(dp: &mut pac::Peripherals) {
    dp.FMC.bcr1.write(|w| w.mbken().disabled());

    dp.FMC.bcr1.write(|w| {
        w.faccen()
            .enabled()
            .mtyp()
            .flash()
            .mwid()
            .bits16()
            .muxen()
            .disabled()
            .mbken()
            .disabled()
            .asyncwait()
            .enabled()
            .cclken()
            .enabled()
            .fmcen()
            .enabled()
    });
    // NOTE unsafe because some values should not be 0
    dp.FMC.btr1.write(|w| unsafe {
        w.addset()
            .bits(9)
            .addhld()
            .bits(1)
            .datast()
            .bits(5)
            .busturn()
            .bits(4)
            .clkdiv()
            .bits(4 - 1)
            .datlat()
            .bits(/* 2 */ 0) // DON'T Care in async mode
            .accmod()
            .b()
    });

    dp.FMC.bcr1.write(|w| {
        w.faccen()
            .enabled()
            .mtyp()
            .flash()
            .mwid()
            .bits16()
            .muxen()
            .disabled()
            .mbken()
            .enabled()
            .asyncwait()
            .enabled()
            .cclken()
            .enabled()
            .fmcen()
            .enabled()
            .wren()
            .enabled()
    })
}

mod nor {
    /// The memory mapped base address for NOR flash
    pub const NOR_ADDR: usize = 0x6000_0000;

    const NOR_CMD_ADDRESS_FIRST: isize = 0x0555;
    const NOR_CMD_ADDRESS_FIRST_CFI: isize = 0x0055;
    const NOR_CMD_ADDRESS_SECOND: isize = 0x02AA;
    const NOR_CMD_ADDRESS_THIRD: isize = 0x0555;
    const NOR_CMD_ADDRESS_FOURTH: isize = 0x0555;
    const NOR_CMD_ADDRESS_FIFTH: isize = 0x02AA;
    const NOR_CMD_ADDRESS_SIXTH: isize = 0x0555;

    const NOR_ADDRESS_COMMAND_SET: isize = 0x0013;

    /* Constants to define data to program a command */
    const NOR_CMD_DATA_READ_RESET: u16 = 0x00F0;
    const NOR_CMD_DATA_FIRST: u16 = 0x00AA;
    const NOR_CMD_DATA_SECOND: u16 = 0x0055;
    const NOR_CMD_DATA_AUTO_SELECT: u16 = 0x0090;
    const NOR_CMD_DATA_PROGRAM: u16 = 0x00A0;
    const NOR_CMD_DATA_CHIP_BLOCK_ERASE_THIRD: u16 = 0x0080;
    const NOR_CMD_DATA_CHIP_BLOCK_ERASE_FOURTH: u16 = 0x00AA;
    const NOR_CMD_DATA_CHIP_BLOCK_ERASE_FIFTH: u16 = 0x0055;
    const NOR_CMD_DATA_CHIP_ERASE: u16 = 0x0010;
    const NOR_CMD_DATA_CFI: u16 = 0x0098;

    const MC_ADDRESS: isize = 0x0000;
    const DEVICE_CODE1_ADDR: isize = 0x0001;
    const DEVICE_CODE2_ADDR: isize = 0x000E;
    const DEVICE_CODE3_ADDR: isize = 0x000F;

    /// Get the command mode from the external flash (16bit or 8bit mode)
    ///
    /// Incase of 16 bit mode, the cmd address must not be shifted by 1,
    /// see https://octopart.com/datasheet/mt28ew128aba1lpc-0sit-micron-61741108
    pub fn get_command_mode(addr: *mut u16) -> u16 {
        unsafe {
            let p = addr.offset(NOR_CMD_ADDRESS_FIRST_CFI);
            core::ptr::write_volatile(p, NOR_CMD_DATA_CFI);
            cortex_m::asm::dsb();
            let p = addr.offset(NOR_ADDRESS_COMMAND_SET);
            let mode = core::ptr::read_volatile(p);
            cortex_m::asm::dsb();
            core::ptr::write_volatile(p, NOR_CMD_DATA_READ_RESET);
            cortex_m::asm::dsb();
            mode
        }
    }

    #[derive(Clone, Debug, Copy)]
    pub enum Error {
        Failed,
        Ongoing,
    }

    /// Get the flash status to check if any error has occured
    pub fn get_status(addr: *mut u16) -> Result<(), Error> {
        const NOR_MASK_STATUS_DQ4: u16 = 0x0010;
        const NOR_MASK_STATUS_DQ5: u16 = 0x0020;
        const NOR_MASK_STATUS_DQ6: u16 = 0x0040;
        const NOR_MASK_STATUS_DQ7: u16 = 0x0080;

        let mut status = Error::Ongoing;
        for _ in 0..10 {
            /* Read NOR status register (DQ6 and DQ5) */
            let t = unsafe {
                (
                    core::ptr::read_volatile(addr),
                    core::ptr::read_volatile(addr),
                )
            };

            /* If DQ6 did not toggle between the two reads then return HAL_NOR_STATUS_SUCCESS  */
            if (t.0 & NOR_MASK_STATUS_DQ6) == (t.1 & NOR_MASK_STATUS_DQ6) {
                return Ok(());
            }

            if t.0 & NOR_MASK_STATUS_DQ5 == NOR_MASK_STATUS_DQ5 {
                status = Error::Ongoing;
            }

            let t = unsafe {
                (
                    core::ptr::read_volatile(addr),
                    core::ptr::read_volatile(addr),
                )
            };

            /* If DQ6 did not toggle between the two reads then return HAL_NOR_STATUS_SUCCESS  */
            if (t.0 & NOR_MASK_STATUS_DQ6) == (t.1 & NOR_MASK_STATUS_DQ6) {
                return Ok(());
            }

            if (t.0 & NOR_MASK_STATUS_DQ5) == NOR_MASK_STATUS_DQ5 {
                return Err(Error::Failed);
            }
        }

        Err(status)
    }

    #[derive(Clone, Copy, Debug, Default)]
    pub struct NorIdCode {
        /// Defines the device's manufacturer code used to identify the memory
        pub manufacturer_code: u16,
        pub device_code: [u16; 3],
    }

    /// Read the manufacturer id to verify, to which type of flash we are talking to.
    pub fn read_id(addr: *mut u16) -> NorIdCode {
        unsafe {
            let p = addr.offset(NOR_CMD_ADDRESS_FIRST);
            core::ptr::write_volatile(p, NOR_CMD_DATA_CFI);
            cortex_m::asm::dsb();

            let id = NorIdCode {
                manufacturer_code: core::ptr::read_volatile(addr.offset(MC_ADDRESS)),
                device_code: [
                    core::ptr::read_volatile(addr.offset(DEVICE_CODE1_ADDR)),
                    core::ptr::read_volatile(addr.offset(DEVICE_CODE2_ADDR)),
                    core::ptr::read_volatile(addr.offset(DEVICE_CODE3_ADDR)),
                ],
            };

            core::ptr::write_volatile(p, NOR_CMD_DATA_READ_RESET);
            cortex_m::asm::dsb();

            id
        }
    }
}
